@@ -10,7 +10,11 @@ expressions in the template.
 '''
 
 
-import collections
+try:
+	import collections.abc as abc
+except ImportError:
+	# python < version 3.3
+	import collections as abc
 
 from zim.utils import MovingWindowIter
 from zim.parser import SimpleTreeElement
@@ -28,7 +32,7 @@ class TemplateContextDict(ExpressionDictObject):
 
 	_fmethods = ExpressionDictObject._fmethods + (
 		'pop', 'clear', 'update', 'setdefault', 'items'
-	) # adding methods for mutuable mapping here
+	) # adding methods for mutable mapping here
 
 
 class TemplateProcessor(object):
@@ -70,9 +74,9 @@ class TemplateProcessor(object):
 
 	def process(self, output, context):
 		'''Execute the template once
-		@param output: an object to recieve the template output, can be
+		@param output: an object to receive the template output, can be
 		a C{list} and should support at least an C{append()} method to
-		recieve string content
+		receive string content
 		@param context: a L{TemplateContextDict} object with the
 		template parameters
 		'''
@@ -146,9 +150,9 @@ class TemplateProcessor(object):
 		var = element.attrib['var']
 		expr = element.attrib['expr']
 		items = expr(context)
-		if not isinstance(items, collections.Iterable):
+		if not isinstance(items, abc.Iterable):
 			raise TypeError('Can not iterate over: %s' % items)
-		elif not isinstance(items, collections.Sized):
+		elif not isinstance(items, abc.Sized):
 			# cast to list to ensure we have a len()
 			items = list(items)
 
